#include <iostream>
#include <fstream>
#include <chrono>

using namespace std;
using namespace std::chrono;

class SudokuRecursive
{
private:
    static const int BLOCK_SIZE = 3;
    static const int GRID_SIZE  = BLOCK_SIZE*BLOCK_SIZE;

    int grid[GRID_SIZE][GRID_SIZE];  // the puzzle grid

    /**
     * Read the template of given numbers and empty cells (zeroes).
     * @param file_name the name of the text file to read.
     */
    void read(char *file_name)
    {
        try
        {
            ifstream grid_file(file_name);
            grid_file.exceptions(ifstream::failbit | ifstream::badbit);

            cout << "Starting template " << file_name << ":" << endl << endl;

            for (int row = 0; row < GRID_SIZE; row++)
            {
                for (int col = 0; col < GRID_SIZE; col++)
                {
                    grid_file >> grid[row][col];
                }
            }
        }
        catch (ifstream::failure& ex)
        {
            cerr << "Input file error: " << file_name << endl;
            exit(-1);
        }
    }

    /**
     * Check if it's OK to assign the number to grid[row][col].
     * @param row the row.
     * @param col the column.
     * @param number the number to assign.
     * @return true if it's OK, else return false.
     */
    bool is_ok(int row, int col, int number)
    {
        // Check that the row doesn't already contain the number.
        for (int c = 0; c < GRID_SIZE; c++)
        {
            if (grid[row][c] == number) return false;  // Not OK.
        }

        // Check that the column doesn't already contain the number.
        for (int r = 0; r < GRID_SIZE; r++)
        {
            if (grid[r][col] == number) return false;  // Not OK.
        }

        // Row and column limits of the block.
        int block_row_start = row - row%BLOCK_SIZE;
        int block_col_start = col - col%BLOCK_SIZE;
        int block_row_end = block_row_start + BLOCK_SIZE;
        int block_col_end = block_col_start + BLOCK_SIZE;

        // Check that the block doesn't already contain the number.
        for (int row = block_row_start; row < block_row_end; row++)
        {
            for (int col = block_col_start; col < block_col_end; col++)
            {
                if (grid[row][col] == number) return false;  // Not OK.
            }
        }

        return true;  // OK to assign the number to the cell.
    }

    /**
     * Solve the puzzle using recursion and backtracking
     * starting at a particular cell.
     * @param row the row of the cell.
     * @param col the column of the cell.
     * @return true if a solution is possible, else return false.
     */
    bool solve(int row, int col)
    {
        // If we're done, stop the recursion.
        if ((row == GRID_SIZE-1) && (col == GRID_SIZE)) return true;

        // Wrap to the next row?
        if (col == GRID_SIZE)
        {
            row++;
            col = 0;
        }

        // If this cell is already filled, recursively solve
        // the rest of the grid starting at the next cell.
        if (grid[row][col] != 0) return solve(row, col+1);

        // Try each number for this cell.
        for (int number = 1; number <= GRID_SIZE; number++)
        {
            if (is_ok(row, col, number))
            {
                // Tentatively assign the number.
                grid[row][col] = number;

                // Can we recursively solve the rest of the grid with
                // that assigned number, starting at the next cell?
                if (solve(row, col+1)) return true;
            }

            // No. Backtrack by removing this number
            //     and then trying the next number.
            grid[row][col] = 0;
        }

        // All the numbers failed for this cell.
        // Return false and the caller will backtrack
        // and try the next number for its cell.
        return false;
    }

public:
    /**
     * Constructor.
     * @param file_name the name of the text file to read.
     */
    SudokuRecursive(char *file_name)
    {
        read(file_name);
    }

    /**
     * Solve the puzzle.
     */
    bool solve()
    {
        return solve(0, 0);
    }

    /**
     * Print the grid with a dot for each empty cell.
     * Include lines to separate the blocks.
     */
    void print()
    {
        for (int row = 0; row < GRID_SIZE; row++)
        {
            if ((row > 0) && (row%BLOCK_SIZE == 0))
            {
                cout << "------+-------+-------" << endl;
            }

            for (int col = 0; col < GRID_SIZE; col++)
            {
                if ((col > 0) && (col%BLOCK_SIZE == 0)) cout << "| ";

                int g = grid[row][col];
                if (g > 0) cout << g << " ";
                else       cout << ". ";
            }

            cout << endl;
        }
    }
};

int main(int argc, char *argv[])
{
    SudokuRecursive sudoku(argv[1]);
    sudoku.print();

    steady_clock::time_point start_time = steady_clock::now();
    bool solved = sudoku.solve();
    steady_clock::time_point end_time = steady_clock::now();

    if (solved)
    {
        cout << endl << "Solution:" << endl << endl;
        sudoku.print();
    }
    else
    {
        cout << endl << "No solution" << endl;
    }

    long us = duration_cast<microseconds>(end_time - start_time).count();
    cout << endl << "Solution time: " << us/1.0e6 << " seconds "<< endl;

    return 0;
}
